import numpy as np
from .utils import EHMNetNode, EHM2NetNode, EHMNet, EHM2Tree, gen_clusters


class EHM:
    """Efficient Hypothesis Management (EHM)

    An implementation of the EHM algorithm, as documented in [EHM1]_.

    """

    @staticmethod
    def construct_net(validation_matrix):
        """Construct the EHM net as per Section 3.1 of [EHM1]_

        Parameters
        ----------
        validation_matrix: :class:`numpy.ndarray`
            An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
            (aka. valid) associations between tracks and detections. The first column corresponds
            to the null hypothesis (hence contains all ones).

        Returns
        -------
        : :class:`~.EHMNet`
            The constructed net object
        """
        num_tracks = validation_matrix.shape[0]

        # Initialise net
        root_node = EHMNetNode(layer=-1)  # Root node is at layer -1
        net = EHMNet([root_node], validation_matrix=validation_matrix)

        # A layer in the network is created for each track (not counting the root-node layer)
        num_layers = num_tracks
        for i in range(num_layers):

            # Get list of nodes in previous layer
            parent_nodes = [node for node in net.nodes if node.layer == i - 1]

            # Get indices of hypothesised detections for the track
            v_detections = set(np.flatnonzero(validation_matrix[i, :]))

            # Compute accumulated measurements up to next layer (i+1)
            acc = set()
            for ii in range(i + 1, num_layers):
                acc |= set(np.flatnonzero(validation_matrix[ii, :]))

            # List of nodes in current layer
            child_nodes = []

            # For all nodes in previous layer
            for parent in parent_nodes:

                # Exclude any detections already considered by parent nodes (always include null)
                v_detections_m1 = (v_detections - parent.identity) | {0}

                # Iterate over valid detections
                for j in v_detections_m1:

                    # Identity
                    identity = acc.intersection(parent.identity | {j}) - {0}

                    # Find valid nodes in current layer that have the same identity
                    v_children = [child for child in child_nodes if identity == child.identity]

                    # If layer is empty or no valid nodes exist, add new node
                    if not len(v_children):
                        # Create new node
                        child = EHMNetNode(layer=i, identity=identity)
                        # Add node to net
                        net.add_node(child, parent, j)
                        # Add node to list of child nodes
                        child_nodes.append(child)
                    else:
                        # Simply add new edge or update existing one
                        for child in v_children:
                            net.add_edge(parent, child, j)
        return net

    @staticmethod
    def compute_association_probabilities(net, likelihood_matrix):
        """Compute the joint association weights, as described in Section 3.3 of [EHM1]_

        Parameters
        ----------
        net: :class:`~.EHMNet`
            A net object representing the valid joint association hypotheses
        likelihood_matrix: :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the unnormalised
            likelihoods for all combinations of tracks and detections. The first column corresponds
            to the null hypothesis.

        Returns
        -------
        :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the normalised
            association probabilities for all combinations of tracks and detecrtons. The first
            column corresponds to the null hypothesis.
        """
        num_tracks, num_detections = likelihood_matrix.shape
        num_nodes = net.num_nodes

        # Compute p_D (Downward-pass) - Eq. (22) of [EHM1]
        p_D = np.zeros((num_nodes, ))
        p_D[0] = 1
        for child in net.nodes[1:]:
            c_i = child.ind
            parents = net.get_parents(child)
            for parent in parents:
                p_i = parent.ind
                ids = list(net.edges[(parent, child)])
                p_D[c_i] += np.sum(likelihood_matrix[child.layer, ids]*p_D[p_i])

        # Compute p_U (Upward-pass) - Eq. (23) of [EHM1]
        p_U = np.zeros((num_nodes, ))
        p_U[-1] = 1
        for parent in reversed(net.nodes[:-1]):
            p_i = parent.ind
            children = net.get_children(parent)
            for child in children:
                c_i = child.ind
                ids = list(net.edges[(parent, child)])
                p_U[p_i] += np.sum(likelihood_matrix[child.layer, ids]*p_U[c_i])

        # Compute p_DT - Eq. (21) of [EHM1]
        p_DT = np.zeros((num_detections, num_nodes))
        for child in net.nodes:
            c_i = child.ind
            # v_edges = {edge: ids for edge, ids in net.edges.items() if edge[1] == child}
            # for edge, ids in v_edges.items():
            #     p_i = edge[0].ind
            #     for j in ids:
            #         p_DT[j, c_i] += p_D[p_i]
            for parent in net.get_parents(child):
                p_i = parent.ind
                for j in net.edges[(parent, child)]:
                    p_DT[j, c_i] += p_D[p_i]

        # Compute p_T - Eq. (20) of [EHM1]
        p_T = np.ones((num_detections, num_nodes))
        p_T[:, 0] = 0
        for node in net.nodes[1:]:
            n_i = node.ind
            for j in range(num_detections):
                p_T[j, n_i] = p_U[n_i]*likelihood_matrix[node.layer, j]*p_DT[j, n_i]

        # Compute association weights - Eq. (15) of [EHM1]
        a_matrix = np.zeros(likelihood_matrix.shape)
        for i in range(num_tracks):
            node_inds = [n_i for n_i, node in enumerate(net.nodes) if node.layer == i]
            for j in range(num_detections):
                a_matrix[i, j] = np.sum(p_T[j, node_inds])
            # Normalise
            a_matrix[i, :] = a_matrix[i, :]/np.sum(a_matrix[i, :])

        return a_matrix

    @classmethod
    def run(cls, validation_matrix, likelihood_matrix):
        """Run EHM to compute and return association probabilities

        Parameters
        ----------
        validation_matrix : :class:`numpy.ndarray`
            An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
            (aka. valid) associations between tracks and detections. The first column corresponds
            to the null hypothesis (hence contains all ones).
        likelihood_matrix: :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the unnormalised
            likelihoods for all combinations of tracks and detections. The first column corresponds
            to the null hypothesis.

        Returns
        -------
        :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the normalised
            association probabilities for all combinations of tracks and detections. The first
            column corresponds to the null hypothesis.
        """

        # Cluster tracks into groups that share common detections
        clusters, missed_tracks = gen_clusters(validation_matrix, likelihood_matrix)

        # Initialise the association probabilities matrix.
        assoc_prob_matrix = np.zeros(likelihood_matrix.shape)
        assoc_prob_matrix[missed_tracks, 0] = 1  # Null hypothesis is certain for missed tracks

        # Perform EHM for each cluster
        for cluster in clusters:

            # Extract track and detection indices
            c_tracks = cluster.tracks
            c_detections = cluster.detections

            # Extract validation and likelihood matrices for cluster
            c_validation_matrix = cluster.validation_matrix
            c_likelihood_matrix = cluster.likelihood_matrix

            # Construct the EHM net
            net = cls.construct_net(c_validation_matrix)

            # Compute the association probabilities
            c_assoc_prob_matrix = cls.compute_association_probabilities(net, c_likelihood_matrix)

            # Map the association probabilities to the main matrix
            for i, track in enumerate(c_tracks):
                assoc_prob_matrix[track, c_detections] = c_assoc_prob_matrix[i, :]

        return assoc_prob_matrix


class EHM2(EHM):
    """ Efficient Hypothesis Management 2 (EHM2)

    An implementation of the EHM2 algorithm, as documented in [EHM2]_.
    """

    @classmethod
    def construct_net(cls, validation_matrix):
        """ Construct the EHM net as per Section 4 of [EHM2]_

        Parameters
        ----------
        validation_matrix: :class:`numpy.ndarray`
            An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
            (aka. valid) associations between tracks and detections. The first column corresponds
            to the null hypothesis (hence contains all ones).

        Returns
        -------
        : :class:`~.EHMNet`
            The constructed net object

        Raises
        ------
        ValueError
            If the provided ``validation_matrix`` is such that tracks can be divided into separate clusters. See
            the :ref:`Note <note1>` below for work-around.


        .. _note1:

        .. note::
            If the provided ``validation_matrix`` is such that tracks can be divided into separate clusters, this
            method will raise a ValueError exception. To work-around this issue, you can use the
            :func:`~pyehm.utils.gen_clusters` function to first generate individual clusters and then generate a net
            for each cluster, as shown below:

            .. code-block:: python

                from pyehm.core import EHM2
                from pyehm.utils import gen_clusters

                validation_matrix = <Your validation matrix>

                clusters, _ = gen_clusters(validation_matrix)

                nets = []
                for cluster in clusters:
                    nets.append(EHM2.construct_net(cluster.validation_matrix)

        """
        num_tracks = validation_matrix.shape[0]

        # Construct tree
        try:
            tree = cls.construct_tree(validation_matrix)
        except ValueError:
            raise ValueError('The provided validation matrix results in multiple clusters of tracks')

        # Initialise net
        root_node = EHM2NetNode(layer=0, track=0, subnet=0)
        net = EHMNet([root_node], validation_matrix=validation_matrix)

        # Recursively construct next layers
        cls._construct_net_layer(net, tree, 1)

        # Compute and cache nodes per track
        for i in range(num_tracks):
            net.nodes_per_track[i] = [node for node in net.nodes if node.track == i]

        return net

    @classmethod
    def _construct_net_layer(cls, net, tree, layer):

        # Get list of nodes in previous layer
        parent_nodes = [node for node in net.nodes if node.layer == layer - 1 and node.subnet == tree.subtree]

        # Get indices of hypothesised detections for the track
        v_detections = set(np.flatnonzero(net.validation_matrix[tree.track, :]))

        # If this is not an end layer
        if tree.children:

            # Process each subtree
            for child_tree in tree.children:

                # Compute accumulated measurements up to next layer (i+1)
                acc = {0} | child_tree.detections

                # List of nodes in current layer
                child_nodes = []

                # For all nodes in previous layer
                for parent in parent_nodes:

                    # Exclude any detections already considered by parent nodes (always include null)
                    v_detections_m1 = (v_detections - parent.identity) | {0}

                    # Iterate over valid detections
                    for j in v_detections_m1:

                        # Identity
                        identity = acc.intersection(parent.identity | {j}) - {0}

                        # Find valid nodes in current layer that have the same identity
                        v_children = [child for child in child_nodes if identity == child.identity]

                        # If layer is empty or no valid nodes exist, add new node
                        if not len(v_children):
                            # Create new node
                            child = EHM2NetNode(layer=layer, subnet=child_tree.subtree, track=child_tree.track,
                                                identity=identity)
                            # Add node to net
                            net.add_node(child, parent, j)
                            # Add node to list of child nodes
                            child_nodes.append(child)
                        else:
                            # Simply add new edge or update existing one
                            for child in v_children:
                                net.add_edge(parent, child, j)
        else:
            # For all nodes in previous layer
            for parent in parent_nodes:

                # Exclude any detections already considered by parent nodes (always include null)
                v_detections_m1 = (v_detections - parent.identity) | {0}

                # Get leaf child, if any
                child = next((node for node in net.nodes if node.layer == layer and node.subnet == tree.subtree), None)

                # Iterate over valid detections
                for j in v_detections_m1:

                    # If layer is empty or no valid node exist, add new node
                    if not child:
                        # Create new node
                        child = EHM2NetNode(layer=layer, subnet=tree.subtree)
                        # Add node to net
                        net.add_node(child, parent, j)
                    else:
                        # Simply add new edge or update existing one
                        net.add_edge(parent, child, j)

        # Create new layers for each sub-tree
        for i, child_tree in enumerate(tree.children):
            cls._construct_net_layer(net, child_tree, layer + 1)

    @staticmethod
    def construct_tree(validation_matrix):
        """ Construct the EHM2 tree as per section 4.3 of [EHM2]_

        Parameters
        ----------
        validation_matrix: :class:`numpy.ndarray`
            An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
            (aka. valid) associations between tracks and detections. The first column corresponds
            to the null hypothesis (hence contains all ones).

        Returns
        -------
        : :class:`~.EHM2Tree`
            The constructed tree object

        Raises
        ------
        ValueError
            If the provided ``validation_matrix`` is such that tracks can be divided into separate clusters. See
            the :ref:`Note <note2>` below for work-around.


        .. _note2:

        .. note::
            If the provided ``validation_matrix`` is such that tracks can be divided into separate clusters, this
            method will raise a ValueError exception. To work-around this issue, you can use the
            :func:`~pyehm.utils.gen_clusters` function to first generate individual clusters and then generate a tree
            for each cluster, as shown below:

            .. code-block:: python

                from pyehm.core import EHM2
                from pyehm.utils import gen_clusters

                validation_matrix = <Your validation matrix>

                clusters, _ = gen_clusters(validation_matrix)

                trees = []
                for cluster in clusters:
                    trees.append(EHM2.construct_tree(cluster.validation_matrix)

        """
        num_tracks = validation_matrix.shape[0]

        trees = []
        last_subtree_index = -1
        for i in reversed(range(num_tracks)):
            # Get indices of hypothesised detections for the track (minus the null hypothesis)
            v_detections = set(np.flatnonzero(validation_matrix[i, :])) - {0}

            matched = []
            for j, tree in enumerate(trees):
                if v_detections.intersection(tree.detections):
                    matched.append(j)

            if matched:
                children = [trees[j] for j in matched]
                detections = set()
                for tree in children:
                    detections |= tree.detections
                detections |= v_detections
                subtree_index = np.max([c.subtree for c in children])
                tree = EHM2Tree(i, children, detections, subtree_index)
                trees = [trees[j] for j in range(len(trees)) if j not in matched]
            else:
                children = []
                last_subtree_index += 1
                tree = EHM2Tree(i, children, v_detections, last_subtree_index)
            trees.append(tree)

        if len(trees) > 1:
            raise ValueError('The provided validation matrix results in multiple clusters of tracks')

        tree = trees[0]

        # Reverse subtree indices
        max_subtree_ind = tree.subtree
        for node in tree.nodes:
            node.subtree = max_subtree_ind - node.subtree

        return tree

    @staticmethod
    def compute_association_probabilities(net, likelihood_matrix):
        """ Compute the joint association weights, as described in Section 4.2 of [EHM2]_

        Parameters
        ----------
        net: :class:`~.EHMNet`
            A net object representing the valid joint association hypotheses
        likelihood_matrix: :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the unnormalised
            likelihoods for all combinations of tracks and detections. The first column corresponds
            to the null hypothesis.

        Returns
        -------
        :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the normalised
            association probabilities for all combinations of tracks and detecrtons. The first
            column corresponds to the null hypothesis.
        """
        num_tracks, num_detections = likelihood_matrix.shape
        num_nodes = net.num_nodes

        nodes_forwards = net.nodes_forward
        nodes_backwards = list(reversed(nodes_forwards))

        # Compute w_B (Backward-pass) - Eq. (47) of [EHM2]
        w_B = np.zeros((num_nodes,))
        for parent in nodes_backwards:
            p_i = parent.ind

            # If parent is a leaf node
            if parent.track is None:
                w_B[p_i] = 1
                continue

            weight = 0
            v_detections = set(np.flatnonzero(net.validation_matrix[parent.track, :])) - parent.identity
            for det_ind in v_detections:
                v_children = net.children_per_detection[(parent, det_ind)]
                weight_det = likelihood_matrix[parent.track, det_ind]
                for child in v_children:
                    c_i = child.ind
                    weight_det *= w_B[c_i]
                weight += weight_det
            w_B[p_i] = weight

        # Compute w_F (Forward-pass) - Eq. (49) of [EHM2]
        w_F = np.zeros((num_nodes,))
        w_F[0] = 1
        for parent in nodes_forwards:
            # Skip the leaf nodes
            if parent.track is None:
                continue
            p_i = parent.ind
            v_detections = set(np.flatnonzero(net.validation_matrix[parent.track, :])) - parent.identity
            for det_ind in v_detections:
                v_children = net.children_per_detection[(parent, det_ind)]
                for child in v_children:
                    if child.track is None:
                        continue
                    c_i = child.ind
                    sibling_inds = [c.ind for c in v_children if not c == child]
                    sibling_weight = np.prod([w_B[s_i] for s_i in sibling_inds])
                    weight = likelihood_matrix[parent.track, det_ind]*w_F[p_i]*sibling_weight
                    w_F[c_i] += weight

        # Compute association probs - Eq. (46) of [EHM2]
        a_matrix = np.zeros(likelihood_matrix.shape)
        for track in range(num_tracks):
            v_detections = set(np.flatnonzero(net.validation_matrix[track, :]))
            for detection in v_detections:
                for parent in net.nodes_per_track[track]:
                    try:
                        v_children = net.children_per_detection[(parent, detection)]
                    except KeyError:
                        continue
                    weight = likelihood_matrix[track, detection] * w_F[parent.ind]
                    for child in v_children:
                        weight *= w_B[child.ind]
                    a_matrix[track, detection] += weight
            a_matrix[track, :] /= np.sum(a_matrix[track, :])

        return a_matrix



#!/usr/bin/env python
# coding: utf-8
"""
Basic Example
=============
"""

import numpy as np

# %%
# Formulating the possible associations between targets and measurements
# ----------------------------------------------------------------------
#
# Both :class:`~.EHM` and :class:`~.EHM2` operate on a ``validation_matrix`` and a ``likelihood_matrix``. The
# ``validation_matrix`` is an indicator matrix that represents the possible associations between different targets
# and measurements, while the ``likelihood_matrix`` contains the respective likelihoods/probabilities of these
# associations. Both matrices have a shape ``(N_T, N_M+1)``, where ``N_T`` is the number of targets and ``N_M`` is
# the numer of measurements.
#
# For example, assume we have the following scenario of 4 targets and 4 measurements (taken from Section 4.4 of
# [EHM2]_):
#
# +---------------+---------------------------+
# | Target index  | Gated measurement indices |
# +===============+===========================+
# | 0             | 0, 1                      |
# +---------------+---------------------------+
# | 1             | 0, 1, 2, 3                |
# +---------------+---------------------------+
# | 2             | 0, 1, 2                   |
# +---------------+---------------------------+
# | 3             | 0, 3, 4                   |
# +---------------+---------------------------+
#
# where the null measurement hypothesis is given the index of 0. Then the ``validation_matrix`` would be a ``(4, 5)``
# numpy array of the following form:

validation_matrix = np.array([[1, 1, 0, 0, 0],   # 0 -> 0,1
                              [1, 1, 1, 1, 0],   # 1 -> 0,1,2,3
                              [1, 1, 1, 0, 0],   # 2 -> 0,1,2
                              [1, 0, 0, 1, 1]])  # 3 -> 0,3,4

# %%
# The ``likelihood_matrix`` is such that each element ``likelihood_matrix[i, j]`` contains the respective likelihood
# of target ``i`` being associated to measurement ``j``. Therefore, based on the above example, the
# ``likelihood_matrix`` could be the following:

likelihood_matrix = np.array([[0.1, 0.9, 0, 0, 0],
                              [0.1, 0.3, 0.2, 0.4, 0],
                              [0.7, 0.1, 0.2, 0, 0],
                              [0.2, 0, 0, 0.75, 0.05]])

# %%
# Computing joint association probabilities
# -----------------------------------------
# Based on the above, we can use :class:`~.EHM` or :class:`~.EHM2` to compute the joint association probabilities
# matrix ``assoc_matrix`` as follows:

from pyehm.core import EHM, EHM2

assoc_matrix_ehm = EHM.run(validation_matrix, likelihood_matrix)
print('assoc_matrix_ehm =\n {}\n'.format(assoc_matrix_ehm))
# or
assoc_matrix_ehm2 = EHM2.run(validation_matrix, likelihood_matrix)
print('assoc_matrix_ehm2 =\n {}'.format(assoc_matrix_ehm2))

# %%
# Note that both :class:`~.EHM` and :class:`~.EHM2` should produce the same results, although :class:`~.EHM2` should,
# in principle, be significantly faster for large numbers of targets and measurements.

# Check if the probability matrices produced by EHM and EHM2 are equal
print(np.allclose(assoc_matrix_ehm, assoc_matrix_ehm2))


import pytest
import numpy as np

from .jdpa_naive import gen_random_scenario, jpda_naive
from ..core import EHM, EHM2


@pytest.mark.parametrize(
    'scenario_name, ehm_class',
    [
        ('scenarioA', EHM),
        ('scenarioB', EHM),
        ('scenarioC', EHM),
        ('scenarioA', EHM2),
        ('scenarioB', EHM2),
        ('scenarioC', EHM2)
    ]
)
def test_ehm_run(scenario_name, ehm_class, scenarios):
    validation_matrix = scenarios[scenario_name]['validation_matrix']
    likelihood_matrix = scenarios[scenario_name]['likelihood_matrix']

    # Read expected results
    expected_result = jpda_naive(validation_matrix, likelihood_matrix)

    # Run EHM
    result = ehm_class.run(validation_matrix, likelihood_matrix)
    assert np.allclose(result, expected_result)


@pytest.mark.parametrize(
    'num_tracks, num_detections',
    [
        (5, 0),
        (0, 5),
        (0, 0),
        (4, 4),
        (8, 9),
    ]
)
def test_ehm_run_random(num_tracks, num_detections):
    for _ in range(10):
        validation_matrix, likelihood_matrix = gen_random_scenario(num_tracks, num_detections)
        expected_result = jpda_naive(validation_matrix, likelihood_matrix)
        ehm1_result = EHM.run(validation_matrix, likelihood_matrix)
        ehm2_result = EHM2.run(validation_matrix, likelihood_matrix)
        assert np.allclose(expected_result, ehm1_result)
        assert np.allclose(expected_result, ehm2_result)



#!/usr/bin/env python
# coding: utf-8
"""
StoneSoup Plugin
================
"""

# %%
# PyEHM includes implementations of `Stone Soup <https://stonesoup.readthedocs.io/en/v0.1b6/index.html>`_ compatible
# Joint Probabilistic Data Association (JPDA) `DataAssociator
# <https://stonesoup.readthedocs.io/en/v0.1b6/stonesoup.dataassociator.html>`_ classes. These are provided under the
# :class:`~.JPDAWithEHM` :class:`~.JPDAWithEHM2` classes, which implement the :class:`~.EHM` and :class:`~.EHM2`
# algorithms, respectively, and can be used as drop-in replacements to the
# :class:`stonesoup.dataassociator.probability.JPDA` data associator.
#
# In this example, we will showcase how the above classes can be utilised using PyEHM's published entry points to
# Stone Soup, and will compare their performance to :class:`stonesoup.dataassociator.probability.JPDA`. It should be
# noted that the code used below is a modified version of the Stone Soup
# `JPDA tutorial <https://stonesoup.readthedocs.io/en/v0.1b6/auto_tutorials/08_JPDATutorial.html>`_

# %%
# Simulate ground truth
# ---------------------
# We simulate three targets moving in the positive x, y cartesian plane (intersecting approximately half-way through
# their transition).

# We then add truth detections with clutter at each time-step.
from datetime import datetime
from datetime import timedelta
from matplotlib import pyplot as plt
import numpy as np
from scipy.stats import uniform

from stonesoup.models.transition.linear import CombinedLinearGaussianTransitionModel, ConstantVelocity
from stonesoup.types.groundtruth import GroundTruthPath, GroundTruthState
from stonesoup.types.detection import TrueDetection
from stonesoup.types.detection import Clutter
from stonesoup.models.measurement.linear import LinearGaussian


np.random.seed(1991)

truths = set()

start_time = datetime.now()
transition_model = CombinedLinearGaussianTransitionModel([ConstantVelocity(0.005),
                                                          ConstantVelocity(0.005)])

truth = GroundTruthPath([GroundTruthState([0, 1, 0, 1], timestamp=start_time)])
for k in range(1, 21):
    truth.append(GroundTruthState(
        transition_model.function(truth[k-1], noise=True, time_interval=timedelta(seconds=1)),
        timestamp=start_time+timedelta(seconds=k)))
truths.add(truth)

truth = GroundTruthPath([GroundTruthState([0, 1, 20, -1], timestamp=start_time)])
for k in range(1, 21):
    truth.append(GroundTruthState(
        transition_model.function(truth[k-1], noise=True, time_interval=timedelta(seconds=1)),
        timestamp=start_time+timedelta(seconds=k)))
truths.add(truth)

truth = GroundTruthPath([GroundTruthState([10, 0, 0, 1], timestamp=start_time)])
for k in range(1, 21):
    truth.append(GroundTruthState(
        transition_model.function(truth[k-1], noise=True, time_interval=timedelta(seconds=1)),
        timestamp=start_time+timedelta(seconds=k)))
truths.add(truth)

# Generate measurements.
all_measurements = []

measurement_model = LinearGaussian(
    ndim_state=4,
    mapping=(0, 2),
    noise_covar=np.array([[0.75, 0],
                          [0, 0.75]])
    )

prob_detect = 0.9  # 90% chance of detection.

for k in range(20):
    measurement_set = set()

    for truth in truths:
        # Generate actual detection from the state with a 10% chance that no detection is received.
        if np.random.rand() <= prob_detect:
            measurement = measurement_model.function(truth[k], noise=True)
            measurement_set.add(TrueDetection(state_vector=measurement,
                                              groundtruth_path=truth,
                                              timestamp=truth[k].timestamp))

        # Generate clutter at this time-step
        truth_x = truth[k].state_vector[0]
        truth_y = truth[k].state_vector[2]
        for _ in range(np.random.randint(10)):
            x = uniform.rvs(truth_x - 10, 20)
            y = uniform.rvs(truth_y - 10, 20)
            measurement_set.add(Clutter(np.array([[x], [y]]), timestamp=truth[k].timestamp))
    all_measurements.append(measurement_set)


# %%
# Tracking components
# -------------------
# In order to compare the 3 data associators noted above, we will be constructing a set of 3 multi-target trackers,
# each of which will make use of a different data associator. Therefore, we now proceed to define the tracking
# components that will constitute the building blocks of the multi-target trackers.

# %%
# Predictor and Updater
# ~~~~~~~~~~~~~~~~~~~~~
# All 3 multi-target trackers will make use of the same pair of predictor and updater:


from stonesoup.predictor.kalman import KalmanPredictor
predictor = KalmanPredictor(transition_model)


from stonesoup.updater.kalman import KalmanUpdater
updater = KalmanUpdater(measurement_model)


# %%
# Hypothesiser
# ~~~~~~~~~~~~
# Similarly, the same hypothesiser will be used by all trackers:


from stonesoup.hypothesiser.probability import PDAHypothesiser
hypothesiser = PDAHypothesiser(predictor=predictor,
                               updater=updater,
                               clutter_spatial_density=0.125,
                               prob_detect=prob_detect)


# %%
# Metrics
# ~~~~~~~
# Below we define some metrics that will enable us to evaluate and compare the tracking performance of the various
# multi-target trackers. More specifically, we will be using the OSPA metric as our basis for comparison:

from stonesoup.metricgenerator.manager import SimpleManager
from stonesoup.metricgenerator.ospametric import OSPAMetric
from stonesoup.measures import Euclidean

ospa_generator = OSPAMetric(c=10, p=1, measure=Euclidean([0, 2]))

# Metric manager for standard JPDA tracker
metric_manager_jpda = SimpleManager([ospa_generator])

# Metric manager for JPDA with EHM tracker
metric_manager_jpda_ehm = SimpleManager([ospa_generator])

# Metric manager for JPDA with EHM2 tracker
metric_manager_jpda_ehm2 = SimpleManager([ospa_generator])


# %%
# Priors
# ~~~~~~
# We assume that the initial positions of the targets are known in advance. Hence, below we define the set of priors
# and initial tracks that will used by each of the trackers


from stonesoup.types.state import GaussianState
from stonesoup.types.track import Track

prior1 = GaussianState([[0], [1], [0], [1]], np.diag([1.5, 0.5, 1.5, 0.5]), timestamp=start_time)
prior2 = GaussianState([[0], [1], [20], [-1]], np.diag([1.5, 0.5, 1.5, 0.5]), timestamp=start_time)
prior3 = GaussianState([[10], [0], [0], [1]], np.diag([1.5, 0.5, 1.5, 0.5]), timestamp=start_time)
priors = {prior1, prior2, prior3}

# %%
# Data Associators
# ~~~~~~~~~~~~~~~~
# We now define the different data associators that we wish to compare:

# Standard JPDA
from stonesoup.dataassociator.probability import JPDA
jpda = JPDA(hypothesiser=hypothesiser)

# JPDA with EHM
from stonesoup.plugins.pyehm import JPDAWithEHM
jpda_ehm = JPDAWithEHM(hypothesiser=hypothesiser)

# JPDA with EHM2
from stonesoup.plugins.pyehm import JPDAWithEHM2
jpda_ehm2 = JPDAWithEHM2(hypothesiser=hypothesiser)



# %%
# Running the trackers
# --------------------
# Below we define a helper function that runs a given tracker configuration, based on the data associator that we wish
# use. This is essentially equivalent to the :class:`stonesoup.tracker.simple.MultiTargetMixtureTracker` recursion, but
# without the track initiation and deletion:


from stonesoup.types.array import StateVectors
from stonesoup.functions import gm_reduce_single
from stonesoup.types.update import GaussianStateUpdate


def run_tracker(data_associator, metric_manager):
    # Track priors
    tracks = {Track([prior]) for prior in priors}

    s = datetime.now()
    for n, measurements in enumerate(all_measurements):
        hypotheses = data_associator.associate(tracks,
                                               measurements,
                                               start_time + timedelta(seconds=n))

        # Loop through each track, performing the association step with weights adjusted according to
        # JPDA.
        for track in tracks:
            track_hypotheses = hypotheses[track]

            posterior_states = []
            posterior_state_weights = []
            for hypothesis in track_hypotheses:
                if not hypothesis:
                    posterior_states.append(hypothesis.prediction)
                else:
                    posterior_state = updater.update(hypothesis)
                    posterior_states.append(posterior_state)
                posterior_state_weights.append(hypothesis.probability)

            means = StateVectors([state.state_vector for state in posterior_states])
            covars = np.stack([state.covar for state in posterior_states], axis=2)
            weights = np.asarray(posterior_state_weights)

            # Reduce mixture of states to one posterior estimate Gaussian.
            post_mean, post_covar = gm_reduce_single(means, covars, weights)

            # Add a Gaussian state approximation to the track.
            track.append(GaussianStateUpdate(
                post_mean, post_covar,
                track_hypotheses,
                track_hypotheses[0].measurement.timestamp))

        metric_manager.add_data(
            truths, tracks, measurements,
            overwrite=False,  # Don't overwrite, instead add above as additional data
        )
    dt = datetime.now() - s
    return tracks, dt

# %%
# Standard JPDA tracker
# ~~~~~~~~~~~~~~~~~~~~~

tracks_jpda, diff_jpda = run_tracker(jpda, metric_manager_jpda)


# %%
# JPDA with EHM tracker
# ~~~~~~~~~~~~~~~~~~~~~

tracks_jpda_ehm, diff_jpda_ehm = run_tracker(jpda_ehm, metric_manager_jpda_ehm)


# %%
# JPDA with EHM2 tracker
# ~~~~~~~~~~~~~~~~~~~~~~

tracks_jpda_ehm2, diff_jpda_ehm2 = run_tracker(jpda_ehm2, metric_manager_jpda_ehm2)

# %%
# Plot the resulting tracks
# -------------------------
multi_fig = plt.figure(figsize=(10, 12))
ax1 = multi_fig.add_subplot(3, 1, 1)
ax1.set_xlabel("$x$")
ax1.set_ylabel("$y$")
ax1.set_ylim(-2, 25)
ax1.set_xlim(-2, 25)
ax1.set_title("Standard JPDA")

ax2 = multi_fig.add_subplot(3, 1, 2)
ax2.set_xlabel("$x$")
ax2.set_ylabel("$y$")
ax2.set_ylim(-2, 25)
ax2.set_xlim(-2, 25)
ax2.set_title("JPDA with EHM")

ax3 = multi_fig.add_subplot(3, 1, 3)
ax3.set_xlabel("$x$")
ax3.set_ylabel("$y$")
ax3.set_ylim(-2, 25)
ax3.set_xlim(-2, 25)
ax3.set_title("JPDA with EHM2")

axes = [ax1, ax2, ax3]

# Plot ground truth.
for truth in truths:
    for ax in axes:
        ax.plot([state.state_vector[0] for state in truth],
                 [state.state_vector[2] for state in truth],
                 linestyle="--",)

# Plot measurements.
for set_ in all_measurements:
    for ax in axes:
        # Plot actual detections.
        ax.scatter([state.state_vector[0] for state in set_ if isinstance(state, TrueDetection)],
                    [state.state_vector[1] for state in set_ if isinstance(state, TrueDetection)],
                    color='g')
        # Plot clutter.
        ax.scatter([state.state_vector[0] for state in set_ if isinstance(state, Clutter)],
                    [state.state_vector[1] for state in set_ if isinstance(state, Clutter)],
                    color='y',
                    marker='2')

tracks_all = [tracks_jpda, tracks_jpda_ehm, tracks_jpda_ehm2]

for ax, tracks in zip(axes, tracks_all):

    for track in tracks:
        # Plot track.
        ax.plot([state.state_vector[0, 0] for state in track[1:]],  # Skip plotting the prior
                 [state.state_vector[2, 0] for state in track[1:]],
                 marker=".")

    # Plot ellipses representing the gaussian estimate state at each update.
    from matplotlib.patches import Ellipse
    for track in tracks:
        for state in track[1:]:  # Skip the prior
            w, v = np.linalg.eig(measurement_model.matrix()@state.covar@measurement_model.matrix().T)
            max_ind = np.argmax(w)
            min_ind = np.argmin(w)
            orient = np.arctan2(v[1, max_ind], v[0, max_ind])
            ellipse = Ellipse(xy=state.state_vector[(0, 2), 0],
                              width=2*np.sqrt(w[max_ind]),
                              height=2*np.sqrt(w[min_ind]),
                              angle=np.rad2deg(orient),
                              alpha=0.2)
            ax.add_artist(ellipse)
multi_fig.tight_layout()

# %%
# Comparison
# ----------
#
# Tracking Performance
# ~~~~~~~~~~~~~~~~~~~~
#
# First, we will compare the tracking performance of the trackers in terms of the generated OSPA metrics. Below we
# proceed to plot the respective metric for each tracker:

# Metrics for standard JPDA
metrics_jpda = metric_manager_jpda.generate_metrics()
ospa_metric_jpda = {metric for metric in metrics_jpda if metric.title == "OSPA distances"}.pop()

# Metrics for JPDA with EHM
metrics_jpda_ehm = metric_manager_jpda_ehm.generate_metrics()
ospa_metric_jpda_ehm = {metric for metric in metrics_jpda_ehm if metric.title == "OSPA distances"}.pop()

# Metrics for JPDA with EHM
metrics_jpda_ehm2 = metric_manager_jpda_ehm2.generate_metrics()
ospa_metric_jpda_ehm2 = {metric for metric in metrics_jpda_ehm2 if metric.title == "OSPA distances"}.pop()

# Plot
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.plot([i.timestamp for i in ospa_metric_jpda.value], [i.value for i in ospa_metric_jpda.value], 'x-', label='Standard JPDA')
ax.plot([i.timestamp for i in ospa_metric_jpda_ehm.value], [i.value for i in ospa_metric_jpda_ehm.value], '*--', label='JPDA with EHM')
ax.plot([i.timestamp for i in ospa_metric_jpda_ehm2.value], [i.value for i in ospa_metric_jpda_ehm2.value], '+--', label='JPDA with EHM2')

ax.set_ylabel("OSPA distance")
ax.tick_params(labelbottom=False)
_ = ax.set_xlabel("Time")
ax.legend()

# %%
# As we can see, the calculated OSPA is identical for the 3 trackers. This is to be expected, since the
# joint association probabilities produced by the :class:`~.EHM` and :class:`~.EHM2` algorithms (used by the
# :class:`~.JPDAWithEHM` and :class:`~.JPDAWithEHM2` classes, respectively) should be identical to
# the standard JPDA (see :ref:`sphx_glr_auto_examples_ehm_vs_ehm2_vs_jpda.py` for more details).

# %%
# Computation times
# ~~~~~~~~~~~~~~~~~

y = ['Standard JPDA', 'JPDA with EHM', 'JPDA with EHM2']
y_pos = [i for i, _ in enumerate(y)]
x = [diff_jpda.total_seconds(), diff_jpda_ehm.total_seconds(), diff_jpda_ehm2.total_seconds()]

# Plot
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.barh(y_pos, x, tick_label=y)
for i, v in enumerate(x):
    if i == 0:
        ax.text(v - 3, i, str(np.around(v, 2)), color='white')
    else:
        ax.text(v + 1, i, str(np.around(v, 2)))
ax.set_ylabel('Tracker')
ax.set_xlabel('Time (s)')
fig.tight_layout()







# -*- coding: utf-8 -*-
import numpy as np
from stonesoup.dataassociator.probability import JPDA
from stonesoup.types.detection import MissedDetection
from stonesoup.types.hypothesis import SingleProbabilityHypothesis
from stonesoup.types.multihypothesis import MultipleHypothesis
from stonesoup.types.numeric import Probability

from ..core import EHM, EHM2


class JPDAWithEHM(JPDA):
    """ Joint Probabilistic Data Association with Efficient Hypothesis Management (EHM)

    This is a faster alternative of the standard :class:`~.JPDA` algorithm, which makes use of
    Efficient Hypothesis Management (EHM) to efficiently compute the joint associations. See
    Maskell et al. (2004) [EHM1]_ for more details.
    """

    def associate(self, tracks, detections, timestamp, **kwargs):
        """Associate tracks and detections

        Parameters
        ----------
        tracks : set of :class:`stonesoup.types.track.Track`
            Tracks which detections will be associated to.
        detections : set of :class:`stonesoup.types.detection.Detection`
            Detections to be associated to tracks.
        timestamp : :class:`datetime.datetime`
            Timestamp to be used for missed detections and to predict to.

        Returns
        -------
        : mapping of :class:`stonesoup.types.track.Track` : :class:`stonesoup.types.hypothesis.Hypothesis`
            Mapping of track to Hypothesis
        """

        # Calculate MultipleHypothesis for each Track over all
        # available Detections
        hypotheses = {
            track: self.hypothesiser.hypothesise(track, detections, timestamp)
            for track in tracks}

        multi_hypotheses = self._compute_multi_hypotheses(tracks, detections, hypotheses, timestamp)

        return multi_hypotheses

    @staticmethod
    def _run_ehm(validation_matrix, likelihood_matrix):
        return EHM.run(validation_matrix, likelihood_matrix)

    @staticmethod
    def _calc_validation_and_likelihood_matrices(tracks, detections, hypotheses):
        """ Compute the validation and likelihood matrices

        Parameters
        ----------
        tracks: list of :class:`stonesoup.types.track.Track`
            Current tracked objects
        detections : list of :class:`stonesoup.types.detection.Detection`
            Retrieved measurements
        hypotheses: dict
            Key value pairs of tracks with associated detections

        Returns
        -------
        :class:`numpy.ndarray`
            An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
            (aka. valid) associations between tracks and detections. The first column corresponds
            to the null hypothesis (hence contains all ones).
        :class:`numpy.ndarray`
            A matrix of shape (num_tracks, num_detections + 1) containing the unnormalised
            likelihoods for all combinations of tracks and detections. The first column corresponds
            to the null hypothesis.
        """

        # Ensure tracks and detections are lists (not sets)
        tracks, detections = list(tracks), list(detections)

        # Construct validation and likelihood matrices
        # Both matrices have shape (num_tracks, num_detections + 1), where the first column
        # corresponds to the null hypothesis.
        num_tracks, num_detections = len(tracks), len(detections)
        likelihood_matrix = np.zeros((num_tracks, num_detections + 1))
        for i, track in enumerate(tracks):
            for hyp in hypotheses[track]:
                if not hyp:
                    likelihood_matrix[i, 0] = hyp.weight
                else:
                    j = next(d_i for d_i, detection in enumerate(detections)
                             if hyp.measurement == detection)
                    likelihood_matrix[i, j + 1] = hyp.weight
        validation_matrix = likelihood_matrix > 0

        return validation_matrix, likelihood_matrix

    @classmethod
    def _compute_multi_hypotheses(cls, tracks, detections, hypotheses, time):

        # Tracks and detections must be in a list so we can keep track of their order
        track_list = list(tracks)
        detection_list = list(detections)

        # Calculate validation and likelihood matrices
        validation_matrix, likelihood_matrix = \
            cls._calc_validation_and_likelihood_matrices(track_list, detection_list, hypotheses)

        # Run EHM
        assoc_prob_matrix = cls._run_ehm(validation_matrix, likelihood_matrix)

        # Calculate MultiMeasurementHypothesis for each Track over all
        # available Detections with probabilities drawn from the association matrix
        new_hypotheses = dict()

        for i, track in enumerate(track_list):

            single_measurement_hypotheses = list()

            # Null measurement hypothesis
            prob_misdetect = Probability(assoc_prob_matrix[i, 0])
            single_measurement_hypotheses.append(
                SingleProbabilityHypothesis(
                    hypotheses[track][0].prediction,
                    MissedDetection(timestamp=time),
                    measurement_prediction=hypotheses[track][0].measurement_prediction,
                    probability=prob_misdetect))

            # True hypotheses
            for hypothesis in hypotheses[track]:
                if not hypothesis:
                    continue

                # Get the detection index
                j = next(d_i+1 for d_i, detection in enumerate(detection_list)
                         if hypothesis.measurement == detection)

                pro_detect_assoc = Probability(assoc_prob_matrix[i, j])
                single_measurement_hypotheses.append(
                    SingleProbabilityHypothesis(
                        hypothesis.prediction,
                        hypothesis.measurement,
                        measurement_prediction=hypothesis.measurement_prediction,
                        probability=pro_detect_assoc))

            new_hypotheses[track] = MultipleHypothesis(single_measurement_hypotheses, True, 1)

        return new_hypotheses


class JPDAWithEHM2(JPDAWithEHM):
    """ Joint Probabilistic Data Association with Efficient Hypothesis Management 2 (EHM2)

    This is an enhanced version of the :class:`~.JPDAWithEHM` algorithm, that makes use of the Efficient Hypothesis
    Management 2 (EHM2) algorithm to efficiently compute the joint associations. See Horridge et al. (2006) [EHM2]_ for
    more details.
    """

    @staticmethod
    def _run_ehm(validation_matrix, likelihood_matrix):
        return EHM2.run(validation_matrix, likelihood_matrix)



# -*- coding: utf-8 -*-
from typing import Union, List, Sequence

import networkx
import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
from networkx.drawing.nx_pydot import graphviz_layout
from networkx.algorithms.components.connected import connected_components


class EHMNetNode:
    """A node in the :class:`~.EHMNet` constructed by :class:`~.EHM`.

    Parameters
    ----------
    layer: :class:`int`
        Index of the network layer in which the node is placed. Since a different layer in the network is built for
        each track, this also represented the index of the track this node relates to.
    identity: :class:`set` of :class:`int`
        The identity of the node. As per Section 3.1 of [EHM1]_, "the identity for each node is an indication of how
        measurement assignments made for tracks already considered affect assignments for tracks remaining to be
        considered".
    """
    def __init__(self, layer, identity=None):
        # Index of the layer (track) in the network
        self.layer = layer
        # Identity of the node
        self.identity = identity if identity else set()
        # Index of the node when added to the network. This is set by the network and
        # should not be edited.
        self.ind = None

    def __repr__(self):
        return 'EHMNetNode(ind={}, layer={}, identity={})'.format(self.ind, self.layer, self.identity)


class EHM2NetNode(EHMNetNode):
    """A node in the :class:`~.EHMNet` constructed by :class:`~.EHM2`.

    Parameters
    ----------
    layer: :class:`int`
        Index of the network layer in which the node is placed.
    track: :class:`int`
        Index of track this node relates to.
    subnet: :class:`int`
        Index of subnet to which the node belongs.
    identity: :class:`set` of :class:`int`
        The identity of the node. As per Section 3.1 of [EHM1]_, "the identity for each node is an indication of how
        measurement assignments made for tracks already considered affect assignments for tracks remaining to be
        considered".
    """
    def __init__(self, layer, track=None, subnet=0, identity=None):
        super().__init__(layer, identity)
        # Index of track this node relates to
        self.track = track
        # Index of subnet the node belongs to
        self.subnet = subnet

    def __repr__(self):
        return 'EHM2NetNode(ind={}, layer={}, track={}, subnet={}, identity={})'.format(self.ind, self.layer,
                                                                                        self.track, self.subnet,
                                                                                        self.identity)


class EHMNet:
    """Represents the nets constructed by :class:`~.EHM` and :class:`~.EHM2`.

    Parameters
    ----------
    nodes: :class:`list` of :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
        The nodes comprising the net.
    validation_matrix: :class:`numpy.ndarray`
        An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
        (aka. valid) associations between tracks and detections. The first column corresponds
        to the null hypothesis (hence contains all ones).
    edges: :class:`dict`
        A dictionary that represents the edges between nodes in the network. The dictionary keys are tuples of the form
        ```(parent, child)```, where ```parent``` and ```child``` are the source and target nodes respectively. The
        values of the dictionary are the measurement indices that describe the parent-child relationship.
    """
    def __init__(self, nodes, validation_matrix, edges=None):

        self._num_layers = 0

        for n_i, node in enumerate(nodes):
            node.ind = n_i
            if isinstance(node, EHM2NetNode):
                if node.layer + 1 > self._num_layers:
                    self._num_layers = node.layer + 1
            else:
                if node.layer + 2 > self._num_layers:
                    self._num_layers = node.layer + 2
        self._nodes = nodes
        self.validation_matrix = validation_matrix
        self.edges = edges if edges is not None else dict()
        self.parents_per_detection = dict()
        self.children_per_detection = dict()
        self.nodes_per_track = dict()

        self._parents = dict()
        self._children = dict()

    @property
    def root(self) -> Union[EHMNetNode, EHM2NetNode]:
        """The root node of the net."""
        return self.nodes[0]

    @property
    def num_nodes(self) -> int:
        """Number of nodes in the net"""
        return len(self._nodes)

    @property
    def num_layers(self) -> int:
        """Number of layers in the net"""
        return self._num_layers

    @property
    def nodes(self) -> Union[List[EHMNetNode], List[EHM2NetNode]]:
        """The nodes comprising the net"""
        return self._nodes

    @property
    def nodes_forward(self) -> Union[Sequence[EHMNetNode], Sequence[EHM2NetNode]]:
        """The net nodes, ordered by increasing layer"""
        return sorted(self.nodes, key=lambda x: x.layer)

    @property
    def nx_graph(self) -> networkx.Graph:
        """A NetworkX representation of the net. Mainly used for plotting the net."""
        g = nx.Graph()
        for child in sorted(self.nodes, key=lambda x: x.layer):
            parents = self.get_parents(child)
            if isinstance(child, EHM2NetNode):
                track = child.track
            else:
                track = child.layer + 1 if child.layer + 2 < self.num_layers else None
            identity = child.identity
            g.add_node(child.ind, track=track, identity=identity)
            for parent in parents:
                label = str(self.edges[(parent, child)]).replace('{', '').replace('}', '')
                g.add_edge(parent.ind, child.ind, detections=label)
        return g

    def add_node(self, node: Union[EHMNetNode, EHM2NetNode], parent: Union[EHMNetNode, EHM2NetNode], detection: int):
        """Add a new node in the network.

        Parameters
        ----------
        node: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The node to be added.
        parent: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The parent of the node.
        detection: :class:`int`
            Index of measurement representing the parent child relationship.
        """
        # Set the node index
        node.ind = len(self.nodes)
        # Add node to graph
        self.nodes.append(node)
        # Create edge from parent to child
        self.edges[(parent, node)] = {detection}
        # Create parent-child-detection look-up
        self.parents_per_detection[(node, detection)] = {parent}
        self._parents[node] = {parent}
        try:
            self.children_per_detection[(parent, detection)].add(node)
        except KeyError:
            self.children_per_detection[(parent, detection)] = {node}
        try:
            self._children[parent].add(node)
        except KeyError:
            self._children[parent] = {node}

        if isinstance(node, EHM2NetNode):
            if node.layer + 1 > self._num_layers:
                self._num_layers = node.layer + 1
        else:
            if node.layer + 2 > self._num_layers:
                self._num_layers = node.layer + 2

    def add_edge(self, parent: Union[EHMNetNode, EHM2NetNode], child: Union[EHMNetNode, EHM2NetNode], detection: int):
        """ Add edge between two nodes, or update an already existing edge by adding the detection to it.

        Parameters
        ----------
        parent: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The parent node, i.e. the source of the edge.
        child: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The child node, i.e. the target of the edge.
        detection: :class:`int`
            Index of measurement representing the parent child relationship.
        """
        try:
            self.edges[(parent, child)].add(detection)
        except KeyError:
            self.edges[(parent, child)] = {detection}
        try:
            self.parents_per_detection[(child, detection)].add(parent)
        except KeyError:
            self.parents_per_detection[(child, detection)] = {parent}
        try:
            self.children_per_detection[(parent, detection)].add(child)
        except KeyError:
            self.children_per_detection[(parent, detection)] = {child}
        try:
            self._children[parent].add(child)
        except KeyError:
            self._children[parent] = {child}
        try:
            self._parents[child].add(parent)
        except KeyError:
            self._parents[child] = {parent}

    def get_parents(self, node: Union[EHMNetNode, EHM2NetNode]) -> Union[Sequence[EHMNetNode], Sequence[EHM2NetNode]]:
        """Get the parents of a node.

        Parameters
        ----------
        node: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The node whose parents should be returned

        Returns
        -------
        :class:`list` of :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            List of parent nodes
        """
        try:
            parents = list(self._parents[node])
        except KeyError:
            parents = []
        return parents  # [edge[0] for edge in self.edges if edge[1] == node]

    def get_children(self, node: Union[EHMNetNode, EHM2NetNode]) -> Union[Sequence[EHMNetNode], Sequence[EHM2NetNode]]:
        """Get the children of a node.

        Parameters
        ----------
        node: :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            The node whose children should be returned

        Returns
        -------
        :class:`list` of :class:`~.EHMNetNode` or :class:`~.EHM2NetNode`
            List of child nodes
        """
        try:
            children = list(self._children[node])
        except KeyError:
            children = []
        return children  # [edge[1] for edge in self.edges if edge[0] == node]

    def plot(self, ax: plt.Axes = None, annotate=True):
        """Plot the net.

        Parameters
        ----------
        ax: :class:`matplotlib.axes.Axes`
            Axes on which to plot the net
        annotate: :class:`bool`
            Flag that dictates whether or not to draw node and edge labels on the plotted net. The default is ``True``
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.gca()
        g = self.nx_graph
        pos = graphviz_layout(g, prog="dot")
        nx.draw(g, pos, ax=ax, node_size=0)
        if annotate:
            labels = dict()
            for n in g.nodes:
                t = g.nodes[n]['track']
                s = str(g.nodes[n]['identity']) if len(g.nodes[n]['identity']) else 'Ø'
                if t is not None:
                    labels[n] = '{{{}, {}}}'.format(t, s)
                else:
                    labels[n] = 'Ø'
            pos_labels = {}
            for node, coords in pos.items():
                pos_labels[node] = (coords[0] + 10, coords[1])
            nx.draw_networkx_labels(g, pos_labels, ax=ax, labels=labels, horizontalalignment='left')
            edge_labels = nx.get_edge_attributes(g, 'detections')
            nx.draw_networkx_edge_labels(g, pos, edge_labels=edge_labels)


class EHM2Tree:
    """Represents the track tree structure generated by :func:`~pyehm.core.EHM2.construct_tree`.

    The :class:`~.EHM2Tree` object represents both a tree as well as the root node in the tree.

    Parameters
    ----------
    track: :class:`int`
        The index of the track represented by the root node of the tree
    children: :class:`list` of :class:`~.EHM2Tree`
        Sub-trees that are children of the current tree
    detections: :class:`set` of :class:`int`
        Set of accumulated detections
    subtree: :class:`int`
        Index of subtree the current tree belongs to.
    """
    def __init__(self, track, children, detections, subtree):
        self.track = track
        self.children = children
        self.detections = detections
        self.subtree = subtree

    @property
    def depth(self) -> int:
        """The depth of the tree"""
        depth = 1
        c_depth = 0
        for child in self.children:
            child_depth = child.depth
            if child_depth > c_depth:
                c_depth = child_depth
        return depth + c_depth

    @property
    def nodes(self) -> List['EHM2Tree']:
        """The nodes/subtrees in the tree"""
        nodes = [self]
        for child in self.children:
            nodes += child.nodes
        return nodes

    @property
    def nx_graph(self) -> networkx.Graph:
        """A NetworkX representation of the tree. Mainly used for plotting the tree."""
        g = nx.Graph()
        return self._traverse_tree_nx(self, g)

    @classmethod
    def _traverse_tree_nx(cls, tree, g, parent=None):
        child = g.number_of_nodes() + 1
        track = tree.track
        detections = tree.detections
        g.add_node(child, track=track, detections=detections)
        if parent:
            g.add_edge(parent, child)
        for sub_tree in tree.children:
            cls._traverse_tree_nx(sub_tree, g, child)
        return g

    def plot(self, ax: plt.Axes = None):
        """Plot the tree.

        Parameters
        ----------
        ax: :class:`matplotlib.axes.Axes`
            Axes on which to plot the tree
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.gca()
        g = self.nx_graph
        pos = graphviz_layout(g, prog="dot")
        nx.draw(g, pos, ax=ax)
        labels = {n: g.nodes[n]['track'] for n in g.nodes}  # if g.nodes[n]['leaf']}
        pos_labels = {}
        for node, coords in pos.items():
            # if g.nodes[node]['leaf']:
            pos_labels[node] = (coords[0], coords[1])
        nx.draw_networkx_labels(g, pos_labels, ax=ax, labels=labels, font_color='white')


class Cluster:
    """A cluster of tracks sharing common detections.

    Parameters
    ----------
    tracks: :class:`list` of `int`
        Indices of tracks in cluster
    detections: :class:`list` of `int`
        Indices of detections in cluster
    validation_matrix: :class:`numpy.ndarray`
        The validation matrix for tracks and detections in the cluster
    likelihood_matrix: :class:`numpy.ndarray`
        The likelihood matrix for tracks and detections in the cluster

    """
    def __init__(self, tracks=None, detections=None, validation_matrix=None, likelihood_matrix=None):
        self.tracks = tracks
        self.detections = detections
        self.validation_matrix = validation_matrix
        self.likelihood_matrix = likelihood_matrix


def gen_clusters(validation_matrix, likelihood_matrix=None):
    """Cluster tracks into groups sharing detections

    Parameters
    ----------
    validation_matrix: :class:`numpy.ndarray`
        An indicator matrix of shape (num_tracks, num_detections + 1) indicating the possible
        (aka. valid) associations between tracks and detections. The first column corresponds
        to the null hypothesis (hence contains all ones).
    likelihood_matrix: :class:`numpy.ndarray`
        A matrix of shape (num_tracks, num_detections + 1) containing the unnormalised
        likelihoods for all combinations of tracks and detections. The first column corresponds
        to the null hypothesis. The default is None, in which case the likelihood matrices of
        the generated clusters will also be None.

    Returns
    -------
    list of :class:`Cluster` objects
        A list of :class:`Cluster` objects, where each cluster contains the indices of the rows
        (tracks) and columns (detections) pertaining to the cluster
    list of int
        A list of row (track) indices that have not been associated to any detections
    """

    # Validation matrix for all detections except null
    validation_matrix_true = validation_matrix[:, 1:]

    # Initiate parameters
    num_tracks, num_detections = np.shape(validation_matrix_true)  # Number of tracks

    # Form clusters of tracks sharing measurements
    missed_tracks = set([i for i in range(num_tracks)])
    clusters = list()

    # List of tracks gated for each detection
    v_lists = [np.flatnonzero(validation_matrix_true[:, detection]) for detection in range(num_detections)]

    # Get clusters of tracks sharing common detections
    G = to_graph(v_lists)
    track_clusters = [t for t in connected_components(G)]

    # Create cluster objects that contain the indices of tracks (rows) and detections (cols)
    for tracks in track_clusters:
        v_detections = {0}
        for track in tracks:
            v_detections |= set(np.flatnonzero(validation_matrix_true[track, :]) + 1)
        # Extract validation and likelihood matrices for cluster
        tracks = sorted(tracks)
        v_detections = sorted(v_detections)
        c_validation_matrix = validation_matrix[tracks, :][:, v_detections]
        if likelihood_matrix is not None:
            c_likelihood_matrix = likelihood_matrix[tracks, :][:, v_detections]
        else:
            c_likelihood_matrix = None
        clusters.append(Cluster(tracks, v_detections, c_validation_matrix, c_likelihood_matrix))

    # Get tracks (rows) that are not associated to any detections
    detected_tracks = set([j for i in track_clusters for j in i])
    missed_tracks = missed_tracks - detected_tracks

    return clusters, list(missed_tracks)


def to_graph(lst):
    G = nx.Graph()
    for part in lst:
        # each sublist is a bunch of nodes
        G.add_nodes_from(part)
        # it also implies a number of edges:
        G.add_edges_from(to_edges(part))
    return G


def to_edges(lst):
    """
        treat `l` as a Graph and return it's edges
        to_edges(['a','b','c','d']) -> [(a,b),(b,c),(c,d)]
    """
    if not len(lst):
        return
    it = iter(lst)
    last = next(it)
    for current in it:
        yield last, current
        last = current



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import re
import os
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../../'))


# -- Project information -----------------------------------------------------

project = 'PyEHM'
copyright = '2021, Lyudmil Vladimirov'
author = 'Lyudmil Vladimirov'

# The full version, including alpha/beta/rc tags
release = '0.1b1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.autosectionlabel',
    'sphinx_gallery.gen_gallery',
]
napoleon_google_docstring = False

autodoc_default_options = {
    # 'members': None,
    'member-order': 'bysource',
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(sys.version_info), None),
    'matplotlib': ('https://matplotlib.org/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'stonesoup': ('https://stonesoup.readthedocs.io/en/latest/', None),
    'networkx': ('https://networkx.org/documentation/stable/', None)
}

sphinx_gallery_conf = {
    'examples_dirs': ['../examples'],
    'gallery_dirs': ['auto_examples'],
    'filename_pattern': re.escape(os.sep),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = ['css/custom.css']


import pytest
import numpy as np

from ..utils import Cluster, gen_clusters


@pytest.mark.parametrize(
    'validation_matrix, val_clusters, val_unnasoc_inds',
    [
        # 3 tracks, no detections
        (
            np.array([[1],
                      [1],
                      [1]]),
            [],
            [0, 1, 2]
        ),
        # 3 tracks, 3 detections
        #  > 2 detected tracks, without shared detections, and 1 undetected track
        (
            np.array([[1, 1, 0, 1],
                      [1, 0, 1, 0],
                      [1, 0, 0, 0]]),
            [Cluster([0], [0, 1, 3]), Cluster([1], [0, 2])],
            [2]
        ),
        # 9 tracks, 7 detections
        #  > 7 detected tracks, with some shared detections, and 2 undetected tracks
        (
            np.array([[1, 1, 0, 1, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 1, 0, 0],
                      [1, 1, 0, 0, 0, 0, 0, 1],
                      [1, 0, 0, 0, 1, 0, 0, 1],
                      [1, 0, 1, 0, 0, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 1, 0],
                      [1, 0, 0, 0, 1, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 0, 0]]),
            [Cluster([0, 2, 3, 7], [0, 1, 3, 4, 7]), Cluster([1, 4], [0, 2, 5]), Cluster([6], [0, 6])],
            [5, 8]
        ),
        # 9 tracks, 7 detections
        #  > 7 detected tracks, with some shared detections, 2 undetected tracks and 2 non-gated detections
        (
            np.array([[1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 0, 1, 0, 0, 0],
                      [1, 1, 0, 0, 0, 0, 0, 0, 1, 0],
                      [1, 0, 0, 0, 1, 0, 0, 0, 1, 0],
                      [1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                      [1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                      [1, 0, 0, 0, 0, 0, 0, 0, 0, 0]]),
            [Cluster([0, 2, 3, 7], [0, 1, 3, 4, 8]), Cluster([1, 4], [0, 2, 6]), Cluster([6], [0, 7])],
            [5, 8]
        ),
    ]
)
def test_clustering(validation_matrix, val_clusters, val_unnasoc_inds):
    # Generate a random likelihood_matrix
    num_tracks, num_detections = validation_matrix.shape
    likelihood_matrix = np.zeros((num_tracks, num_detections))
    for i in range(num_tracks):
        liks = np.random.rand(num_detections)
        likelihood_matrix[i, :] = validation_matrix[i, :] * liks

    clusters, unnasoc_inds = gen_clusters(validation_matrix, likelihood_matrix)
    assert len(clusters) == len(val_clusters)
    for cluster, val_cluster in zip(clusters, val_clusters):
        assert cluster.tracks == val_cluster.tracks
        assert cluster.detections == val_cluster.detections
        assert np.array_equal(cluster.validation_matrix, validation_matrix[cluster.tracks, :][:, cluster.detections])
        assert np.array_equal(cluster.likelihood_matrix, likelihood_matrix[cluster.tracks, :][:, cluster.detections])
    assert set(unnasoc_inds) == set(val_unnasoc_inds)



import itertools
import numpy as np


def jpda_naive(validation_matrix, likelihood_matrix):
    num_tracks, num_detections = validation_matrix.shape

    possible_assoc = list()
    for track in range(num_tracks):
        track_possible_assoc = list()
        v_detections = np.flatnonzero(validation_matrix[track, :])
        for detection in v_detections:
            track_possible_assoc.append((track, detection))
        possible_assoc.append(track_possible_assoc)

    # Compute all possible joint hypotheses
    joint_hyps = itertools.product(*possible_assoc)

    # Compute valid joint hypotheses
    valid_joint_hypotheses = (joint_hypothesis for joint_hypothesis in joint_hyps if isvalidhyp(joint_hypothesis))

    # Compute likelihood for valid joint hypotheses
    valid_joint_hypotheses_lik = dict()
    for joint_hyp in valid_joint_hypotheses:
        lik = 1
        for hyp in joint_hyp:
            track = hyp[0]
            detection = hyp[1]
            lik *= likelihood_matrix[track, detection]
        valid_joint_hypotheses_lik[joint_hyp] = lik

    assoc_matrix = np.zeros((num_tracks, num_detections))
    for track in range(num_tracks):
        v_detections = np.flatnonzero(validation_matrix[track, :])
        for detection in v_detections:
            prob = np.sum([lik for hyp, lik in valid_joint_hypotheses_lik.items() if (track, detection) in hyp])
            assoc_matrix[track, detection] = prob
        assoc_matrix[track, :] /= np.sum(assoc_matrix[track, :])

    return assoc_matrix


def isvalidhyp(joint_hyp):
    used_detections = set()
    for hyp in joint_hyp:
        detection = hyp[1]
        if not detection:
            pass
        elif detection in used_detections:
            return False
        else:
            used_detections.add(detection)
    return True


def gen_random_scenario(num_tracks, num_detections):

    validation_matrix = np.zeros((num_tracks, num_detections+1))
    validation_matrix[:, 0] = 1
    for i in range(num_tracks):
        validation_matrix[i, 1:] = np.random.randint(0, 2, (num_detections,))

    likelihood_matrix = np.zeros((num_tracks, num_detections + 1))
    for i in range(num_tracks):
        liks = np.random.rand(num_detections+1)
        likelihood_matrix[i, :] = validation_matrix[i, :]*liks

    return validation_matrix, likelihood_matrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from pyehm import __version__ as version

with open('README.md') as f:
    long_description = f.read()

# Setting up
setup(
    name="pyehm",
    version=version,
    author="Lyudmil Vladimirov",
    author_email="sglvladi@liverpool.ac.uk",
    maintainer="University of Liverpool",
    url='https://github.com/sglvladi/pyehm',
    description='Python Efficient Hypothesis Management (PyEHM)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    project_urls={
        'Documentation': 'https://pyehm.rtfd.io/',
        'Source': 'https://github.com/sglvladi/pyehm',
        'Issue Tracker': 'https://github.com/sglvladi/pyehm/issues'
    },
    packages=find_packages(exclude=('docs', '*.tests')),
    setup_requires=['setuptools_scm', 'setuptools_scm_git_archive'],
    install_requires=['numpy', 'networkx', 'stonesoup', 'setuptools>=42', 'pydot', 'matplotlib'],
    extras_require={
        'dev': ['pytest-flake8', 'pytest-cov', 'sphinx', 'sphinx_rtd_theme', 'sphinx-gallery>=0.8']
    },
    entry_points={'stonesoup.plugins': 'pyehm = pyehm.plugins.stonesoup'},
    python_requires='>=3.6',
    keywords=['python', 'pyehm', 'ehm'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ]
)



#!/usr/bin/env python
# coding: utf-8
"""
Standard JPDA vs EHM vs EHM2
============================
"""

# %%
# Both :class:`~.EHM` and :class:`~.EHM2` provide an exact solution to the problem posed by the Joint Probabilistic
# Data Association (JPDA) algorithm. However, even though in the naive implementation JPDA the number of hypotheses,
# and as a direct consequence the time required to evaluate these, increases exponentially with number of targets
# and measurements, :class:`~.EHM` and :class:`~.EHM2` produce results in sub-exponential time.
#
#
# Problem formulation
# -------------------
#
# In this example we will be comparing the computational performance of the :class:`~.EHM` and :class:`~.EHM2`, against
# a naive implementation of JPDA, for a relatively dense scenario of 11 targets and 9 measurements. The
# validation and likelihood matrices for this scenario are defined below (For more information on how these matrices
# are defined, see the :ref:`sphx_glr_auto_examples_basic.py`):

import itertools
import datetime
import numpy as np

from pyehm.core import EHM, EHM2

validation_matrix = np.array([[1, 1, 1, 0, 1, 0, 1, 1, 0, 0],   # 0  -> 0,1,2,4,6,7
                              [1, 1, 0, 1, 1, 1, 1, 1, 0, 0],   # 1  -> 0,1,3,4,5,6,7
                              [1, 1, 0, 1, 0, 1, 1, 1, 1, 0],   # 2  -> 0,1,3,5,6,7,8
                              [1, 1, 1, 1, 0, 0, 1, 1, 0, 1],   # 3  -> 0,1,2,3,6,7,9
                              [1, 0, 1, 1, 0, 0, 0, 0, 1, 0],   # 4  -> 0,2,3,8
                              [1, 1, 1, 0, 0, 1, 1, 1, 1, 0],   # 5  -> 0,1,2,5,6,7,8
                              [1, 1, 0, 0, 0, 1, 1, 0, 1, 1],   # 6  -> 0,1,5,6,8,9
                              [1, 1, 1, 0, 0, 0, 0, 0, 0, 0],   # 7  -> 0,1,2
                              [1, 0, 0, 1, 1, 1, 1, 1, 1, 0],   # 8  -> 0,3,4,5,6,7,8
                              [1, 0, 0, 0, 1, 0, 1, 1, 1, 0],   # 9  -> 0,4,6,7,8
                              [1, 0, 1, 0, 0, 0, 0, 0, 0, 1]])  # 10 -> 0,2,9

likelihood_matrix = np.array([[0.9, 0.13, 0.1, 0, 0.97, 0, 0.94, 0.55, 0, 0],
                              [0.55, 0.31, 0, 0.61, 0.27, 0.38, 0.34, 0.58, 0, 0],
                              [0.61, 0.55, 0, 0.32, 0, 0.25, 0.8, 0.94, 0.62, 0],
                              [0.45, 0.53, 0.61, 0.19, 0, 0, 0.95, 0.61, 0, 0.17],
                              [0.67, 0, 0.79, 0.99, 0, 0, 0, 0, 0.71, 0],
                              [0.51, 0.37, 0.04, 0, 0, 0.53, 0.92, 0.44, 0.95, 0],
                              [0.31, 0.03, 0, 0, 0, 0.08, 0.68, 0, 0.04, 0.31],
                              [0.23, 0.09, 0.21, 0, 0, 0, 0, 0, 0, 0],
                              [0.62, 0, 0, 0.19, 0.17, 0.31, 0.69, 0.89, 0.63, 0],
                              [0.44, 0, 0, 0, 0.53, 0, 0.49, 0.01, 0.31, 0],
                              [0.32, 0, 0.56, 0, 0, 0, 0, 0, 0, 0.23]])
# %%
# EHM vs EHM 2
# ------------
#
# It is worth noticing that in the above example, targets 7 and 10 are conditionally independent of targets 8 and 9,
# given target 6. This is because targets 7 and 10 share a common measurement (2), but do not share any
# measurements with targets 8 or 9, which in turn have common measurements (4, 6, 7, 8). Yet, all of them
# share measurements with target 6.
#
# This is important since :class:`~.EHM2` takes advantage of this conditional independence to reduce the number of
# nodes in the constructed net and, as a result, achieve better computational performance than :class:`~.EHM`.
#
# To better understand the above, let us examine the number of nodes in the nets produced by the two algorithms:
#

# Net constructed using EHM
net1 = EHM.construct_net(validation_matrix)

# Net constructed using EHM2
net2 = EHM2.construct_net(validation_matrix)

print('No. of nodes in EHM net: {}'.format(net1.num_nodes))
print('No. of nodes in EHM2 net: {}'.format(net2.num_nodes))


# %%
# Standard JPDA
# -------------
#
# Below we define the function ``jpda`` that computes the joint association probabilities based on the standard
# JPDA recursion, which performs a full enumeration of all the joint hypotheses.

def jpda(validation_matrix, likelihood_matrix):
    num_tracks, num_detections = validation_matrix.shape

    possible_assoc = list()
    for track in range(num_tracks):
        track_possible_assoc = list()
        v_detections = np.flatnonzero(validation_matrix[track, :])
        for detection in v_detections:
            track_possible_assoc.append((track, detection))
        possible_assoc.append(track_possible_assoc)

    # Compute all possible joint hypotheses
    joint_hyps = itertools.product(*possible_assoc)

    # Compute valid joint hypotheses
    valid_joint_hypotheses = (joint_hypothesis for joint_hypothesis in joint_hyps if is_valid_hyp(joint_hypothesis))

    # Compute likelihood for valid joint hypotheses
    valid_joint_hypotheses_lik = dict()
    for joint_hyp in valid_joint_hypotheses:
        lik = 1
        # The likelihood of a joint hypothesis is the product of the likelihoods of its member hypotheses
        for hyp in joint_hyp:
            track = hyp[0]
            detection = hyp[1]
            lik *= likelihood_matrix[track, detection]
        valid_joint_hypotheses_lik[joint_hyp] = lik

    # Compute the joint association probabilities
    assoc_matrix = np.zeros((num_tracks, num_detections))
    for track in range(num_tracks):
        v_detections = np.flatnonzero(validation_matrix[track, :])
        for detection in v_detections:
            # The joint assoc. probability for a track-detection hypothesis is the sum of the likelihoods of all
            # joint hypotheses that include this hypothesis
            prob = np.sum([lik for hyp, lik in valid_joint_hypotheses_lik.items() if (track, detection) in hyp])
            assoc_matrix[track, detection] = prob
        # Normalise
        assoc_matrix[track, :] /= np.sum(assoc_matrix[track, :])

    return assoc_matrix


def is_valid_hyp(joint_hyp):
    used_detections = set()
    for hyp in joint_hyp:
        detection = hyp[1]
        if not detection:
            pass
        elif detection in used_detections:
            return False
        else:
            used_detections.add(detection)
    return True

# %%
# Comparison
# ----------
#
# Now we can compare the above against :class:`~.EHM` and :class:`~.EHM2`, both in terms of accuracy and computation
# time. The accuracy comparison is just a safe-guard check to make sure that :class:`~.EHM` and :class:`~.EHM2`
# produce the same result as the standard JPDA.

# EHM
now = datetime.datetime.now()
assoc_matrix_ehm = EHM.run(validation_matrix, likelihood_matrix)
dt_ehm = datetime.datetime.now() - now

# EHM2
now = datetime.datetime.now()
assoc_matrix_ehm2 = EHM2.run(validation_matrix, likelihood_matrix)
dt_ehm2 = datetime.datetime.now() - now

# Standard JPDA
now = datetime.datetime.now()
assoc_matrix_jpda = jpda(validation_matrix, likelihood_matrix)
dt_jpda = datetime.datetime.now() - now

# Check if all results are the same
print(np.allclose(assoc_matrix_jpda, assoc_matrix_ehm, atol=1e-15)
      and np.allclose(assoc_matrix_jpda, assoc_matrix_ehm2, atol=1e-15))

# Compare the execution times
print('JPDA: {} seconds'.format(dt_jpda.total_seconds()))
print('EHM: {} seconds'.format(dt_ehm.total_seconds()))
print('EHM2: {} seconds'.format(dt_ehm2.total_seconds()))

# %%
# The above results demonstrate the advantages of using the :class:`~.EHM` and :class:`~.EHM2` classes over the
# standard JPDA. Both the :class:`~.EHM` and :class:`~.EHM2` algorithms exhibit significant computational gains
# compared to the standard JPDA, all while producing exactly the same results. We can also observe that :class:`~.EHM2`
# is noticeably faster than :class:`~.EHM`.
